<?php

namespace Drupal\kinetic_extras\Generators;

use Drupal\Core\Asset\LibraryDiscoveryInterface;
use DrupalCodeGenerator\Asset\File;
use DrupalCodeGenerator\Command\DrupalGenerator;
use Symfony\Component\Console\Question\Question;

/**
 * Integration service with the drush generate framework.
 */
final class ComponentGenerator extends DrupalGenerator {

  /**
   * {@inheritdoc}
   */
  protected ?string $nameQuestion = NULL;
  /**
   * {@inheritdoc}
   */
  protected ?int $extensionType = self::EXTENSION_TYPE_THEME;
  /**
   * {@inheritdoc}
   */
  protected bool $isNewExtension = FALSE;
  /**
   * {@inheritdoc}
   */
  protected string $name = 'kinetic:sdc';
  /**
   * {@inheritdoc}
   */
  protected string $description = 'Generates an empty component.';
  /**
   * {@inheritdoc}
   */
  protected string $alias = 'kc';
  /**
   * {@inheritdoc}
   */
  protected string $templatePath = __DIR__;

  /**
   * Illustrates how to inject a dependency into a Generator.
   *
   * @var \Drupal\Core\Asset\LibraryDiscoveryInterface
   */
  protected LibraryDiscoveryInterface $libraryDiscovery;

  /**
   * Creates a new generator object.
   *
   * @param \Drupal\Core\Asset\LibraryDiscoveryInterface $library_discovery
   *   Library discovery.
   */
  public function __construct(LibraryDiscoveryInterface $library_discovery) {
    parent::__construct($this->name);
    $this->libraryDiscovery = $library_discovery;
  }

  /**
   * {@inheritdoc}
   */
  protected function generate(&$vars): void {
    $vars = $this->askQuestions($vars);
    $this->generateAssets($vars);
  }

  /**
   * Asks the questions to the user.
   *
   * @param array $vars
   *   The answers array passed by reference.
   *
   * @return array
   *   The answers to the questions.
   */
  private function askQuestions(&$vars): array {
    $this->setUpGeneralInfo($vars);
    $this->setUpDestination($vars);
    $this->setupName($vars);
    $this->setupDescription($vars);
    $this->setupStatus($vars);
    $this->setupComponentLibraries($vars);

    $vars['component_has_css'] = $this->confirm('Needs CSS?');
    $vars['component_has_js'] = $this->confirm('Needs JS?');

    $this->selectEntityTypeHookup($vars);
    $this->handleEntityTypeTemplateName($vars);

    $this->setupComponentProps($vars);

    return $vars;
  }

  /**
   * Sets up general information.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setUpGeneralInfo(&$vars) {
    $extension_type = 'theme';
    $this->extensionType = $extension_type === 'module' ? self::EXTENSION_TYPE_MODULE : self::EXTENSION_TYPE_THEME;
    $this->machineNameQuestion = 'Machine name of the ' . $extension_type;
    $vars['machine_name'] = 'kinetic';
    $vars['name'] = 'Kinetic';
  }

  /**
   * Sets up the destination directory.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setUpDestination(&$vars) {
    $destination_choices = [
      'source/02-components/00-elements' => 'Elements',
      'source/02-components/01-composites' => 'Composites',
      'source/02-components/02-blocks' => 'Blocks',
      'source/02-components/03-formations' => 'Formations',
      'source/02-components/04-pages' => 'Pages',
    ];
    $vars['directory'] = $this->choice('Choose destination', $destination_choices);
  }

  /**
   * Sets up the component name and machine name.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setupName(&$vars) {
    $vars['component_machine_name'] = $this->ask('Component machine name', 'my-component', '::validateRequiredMachineName');

    $default = \ucwords(\trim(\str_replace([
      '_',
      '-',
    ], ' ', $vars['component_machine_name'])));

    $vars['component_name'] = $this->ask('Component name', $default, '::validateRequired');
  }

  /**
   * Sets up the component description.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setupDescription(&$vars) {
    $vars['component_description'] = $this->ask('Component description (optional)');
  }

  /**
   * Sets up the component status.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setupStatus(&$vars) {
    $status_choices = [
      "stable" => 'Stable',
      "experimental" => 'Experimental',
      "deprecated"  => 'Deprecated',
      "obsolete" => 'Obsolete',
    ];
    $vars['component_status'] = $this->choice('Status', $status_choices);
  }

  /**
   * Sets up the component libraries.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setupComponentLibraries(&$vars) {
    $vars['component_libraries'] = [];
    do {
      $library = $this->askLibrary();
      $vars['component_libraries'][] = $library;
    } while (!empty($library));
    $vars['component_libraries'] = \array_filter($vars['component_libraries']);
  }

  /**
   * Sets up the component props.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function setupComponentProps(&$vars) {
    if ($this->confirm('Needs component props?')) {
      $vars['component_props'] = [];
      do {
        $prop = $this->askProp();
        $vars['component_props'][] = $prop;
      } while ($this->confirm('Add another prop?'));
    }
    $vars['component_props'] = \array_filter($vars['component_props'] ?? []);
  }

  private function selectEntityTypeHookup(&$vars) {
    $choices = [
      'none' => 'None',
      'block' => 'Block',
      'paragraph' => 'Paragraph',
//      'taxonomy' => 'Taxonomy',
//      'media' => 'Media',
//      'node' => 'Node',
    ];
    $vars['entityType'] = $this->choice('Stub a entity template?', $choices, 'none');
  }

  // handle the entity type template naming if entity type is selected.
  private function handleEntityTypeTemplateName(&$vars) {
    if ($vars['entityType']) {
      if ($vars['entity_machine_name']) {
        $name_suggestion = $vars['entity_machine_name'];
      }
      else {
        $name_suggestion = $vars['component_machine_name'];
      }
      $full_name_suggestion = $vars['entityType'] . '--' . $name_suggestion;
      $vars['entity_template_name'] = $this->ask('Entity template name (will add .html.twig for you)', $full_name_suggestion, '::validateRequiredMachineName');
    }
  }

  /**
   * Create the assets that the framework will write to disk later on.
   *
   * @param array $vars
   *   The answers to the CLI questions.
   */
  private function generateAssets(array $vars): void {
    $component_path_token = '{directory}' . DIRECTORY_SEPARATOR . '{component_machine_name}' . DIRECTORY_SEPARATOR;

    $this->generateCss($vars, $component_path_token);
    $this->generateJs($vars, $component_path_token);
    $this->generateSdcTemplate($vars, $component_path_token);
    $this->generateComponentYml($vars, $component_path_token);
    $this->generateComponentMisc($vars, $component_path_token);
    //    entityType
    //    entity_machine_name
    //    create_entity
    //    create_layout_builder_browser_entry
    //    block_category
    $entity_path_token = $this->createEntityPathToken($vars);
    $this->createEntityTemplate($vars, $entity_path_token);
  }

  // handle css generation.
  private function generateCss(array $vars, $component_path_token): void {
    if ($vars['component_has_css']) {
      $this->addFile(
        $component_path_token . '{component_machine_name}.scss',
        'main-css--template.twig'
      );
    }
  }

  // handle js generation.
  private function generateJs(array $vars, $component_path_token): void {
    if ($vars['component_has_js']) {
      $this->addFile(
        $component_path_token . '{component_machine_name}.es6.js',
        'main-js--template.twig'
      );
    }
  }

  private function generateSdcTemplate(array $vars, $component_path_token): void {
    $this->addFile(
      $component_path_token . '{component_machine_name}.twig',
      'component-twig--template.twig'
    );
  }

  private function generateComponentYml(array $vars, $component_path_token): void {
    $this->addFile($component_path_token . '{component_machine_name}.component.yml', 'component-yml--template.twig');
  }

  private function generateComponentMisc(array $vars, $component_path_token): void {
    $this->addFile($component_path_token . 'README.md', 'readme-md--template.twig');
    $contents = file_get_contents(__DIR__ . DIRECTORY_SEPARATOR . 'thumbnail-placeholder.png');
    $thumbnail = new File($component_path_token . 'thumbnail.png');
    $thumbnail->content($contents);
    $this->assets[] = $thumbnail;
  }

  // handle entity template generation.
  private function createEntityPathToken(array $vars): string {
    if ($vars['entityType']) {
      if ($vars['entityType'] === 'block') {
        $entity_path_token = 'templates/' . $vars['entityType'] . '/' . '{component_machine_name}' . DIRECTORY_SEPARATOR;
      }
      else {
        // for other entity types (paragraph, taxonomy, media, node
        $entity_path_token = 'templates/' . $vars['entityType'] . DIRECTORY_SEPARATOR;
      }
    }
    return $entity_path_token;
  }

  // handle entity template generation.
  private function createEntityTemplate(array $vars, $entity_path_token): void {
    if ($vars['entityType']) {
      // add a switch statement to handle the different entity types.
      switch ($vars['entityType']) {
        case 'block':
          $this->addFile($entity_path_token . $vars['entity_template_name'] . '.html.twig', 'entity--block-html-twig--template.twig');
          break;
        case 'paragraph':
          $this->addFile($entity_path_token . $vars['entity_template_name'] . '.html.twig', 'entity--paragraph-html-twig--template.twig');
          break;
//        case 'taxonomy':
//          $this->addFile($entity_path_token . $vars['entityTemplateName'] . '.html.twig', 'entity--taxonomy-html-twig--template.twig');
//          break;
//        case 'media':
//          $this->addFile($entity_path_token . $vars['entityTemplateName'] . '.html.twig', 'entity--media-html-twig--template.twig');
//          break;
//        case 'node':
//          $this->addFile($entity_path_token . $vars['entityTemplateName'] . '.html.twig', 'entity--node-html-twig--template.twig');
//          break;
      }
    }
  }

  /**
   * Prompts the user for a library.
   *
   * This helper gathers all the libraries from the system to allow autocomplete
   * and validation.
   *
   * @return string|null
   *   The library ID, if any.
   */
  private function askLibrary(): ?string {
    $library_question = new Question('Library dependencies (optional). [Example: core/once]');
    $extensions = [
      'core',
      ...\array_keys($this->drupalContext->getModules()),
      ...\array_keys($this->drupalContext->getThemes()),
    ];
    $library_ids = array_reduce(
      $extensions,
      fn(array $libs, string $extension) => \array_merge(
        $libs,
        \array_map(static fn (string $l) => sprintf('%s/%s', $extension, $l),
          \array_keys($this->libraryDiscovery->getLibrariesByExtension($extension)))
      ),
      []
    );
    $library_question->setValidator(static::libraryValidator($library_ids));
    $library_question->setAutocompleterValues($library_ids);
    return $this->io->askQuestion($library_question);
  }

  /**
   * Asks for multiple questions to define a prop and its schema.
   *
   * @return array
   *   The prop data, if any.
   */
  protected function askProp(): array {
    $prop = [];
    $prop['name'] = $this->ask('Prop machine name', NULL, '::validateRequiredMachineName');
    $default = str_replace(['-', '_'], ' ', $prop['name']);
    $default = ucwords($default);
    $prop['title'] = $this->ask('Prop human name', $default);
    $prop['description'] = $this->ask('Prop description (optional)');
    $prop['type'] = $this->choice('Prop type', [
      'string' => 'String',
      'number' => 'Number',
      'boolean' => 'Boolean',
      'array' => 'Array',
      'object' => 'Object',
      'null' => 'Always null',
    ]);
    if (!in_array($prop['type'], ['string', 'number', 'boolean'])) {
      $this->io->warning('Unable to generate full schema for ' . $prop['type'] . '. Please edit metadata.json after generation.');
      return $prop;
    }
    $prop['examples'] = [];
    if ($prop['type'] === 'string') {
      // Gather examples.
      do {
        $example = $this->ask('Give a prop example (optional)');
        $prop['examples'][] = $example;
      } while (!is_null($example));
      $prop['examples'] = \array_filter($prop['examples']);
    }
    return $prop;
  }

  /**
   * Builds the library validator based on the allowed list.
   *
   * @param array $library_ids
   *   Allowed libraries.
   *
   * @return callable
   *   The validator function.
   */
  public static function libraryValidator(array $library_ids): callable {
    return static fn($value) => is_null($value) || \in_array($value, $library_ids, TRUE) ? $value : throw new \UnexpectedValueException('Invalid library ID');
  }

  /**
   * Validates that a string is in kebab case.
   *
   * @param string|null $value
   *   The input value to validate.
   */
  public static function validateMachineName(?string $value): string {
    return static::validate(
      $value,
      '^[a-z][a-z0-9-]*[a-z0-9]$',
      'The value is not in the correct format. Ex: a-kebab-case-str1ng'
    ) ?? '';
  }

  /**
   * Validates that a string is in kebab case and it is required.
   *
   * @param string|null $value
   *   The input value to validate.
   */
  public static function validateRequiredMachineName(?string $value): string {
    $value = static::validateRequired($value);
    return static::validate(
      $value,
      '^[a-z][a-z0-9-_]*[a-z0-9]$',
      'The value is not in the correct format. Ex: a-kebab-case-str1ng'
    );
  }

}
